// Copyright  2004 by Joe White (joe@excastle.com).
// Licensed under the Artistic License. See Artistic.txt.

using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Text;
using System.Windows.Forms;

namespace SharpLayout
{
	public class FlowLabel : Control
	{
		private bool _antiAlias;
		private Color _borderColor = SystemColors.ControlDark;
		private Borders _borders;
		private int _imageIndex = -1;
		private ImageList _imageList;
		private int _imagePadding = 4;
		private Padding _padding;
		private bool _underbar;

		public FlowLabel()
		{
			_borders = new Borders();
			_borders.Changed += new EventHandler(BordersChanged);
			_padding = new Padding();
			_padding.Changed += new EventHandler(PaddingChanged);
			SetStyle(ControlStyles.DoubleBuffer | ControlStyles.AllPaintingInWmPaint, true);
		}


		private Padding AllPadding
		{
			get
			{
				Padding result = Padding.Clone();

				if (Borders.Left)
					result.Left += BorderWidth;
				if (Borders.Top)
					result.Top += BorderWidth;
				if (Borders.Right)
					result.Right += BorderWidth;
				if (Borders.Bottom)
					result.Bottom += BorderWidth;

				if (Underbar)
					result.Bottom += TotalUnderbarHeight;

				return result;
			}
		}

		private Padding AllTextPadding
		{
			get
			{
				Padding allPadding = AllPadding;

				Padding padding = AllPadding.Clone();

				// Center the first line of text vertically within the image
				if (ImageList != null)
				{
					padding.Left += ImageList.ImageSize.Width + ImagePadding;

					int lineHeight = Font.Height;
					int imageHeight = ImageList.ImageSize.Height;
					if (imageHeight > lineHeight)
						padding.Top += (imageHeight - lineHeight) / 2;
				}
				return padding;
			}
		}

		[DefaultValue(false)]
		public bool AntiAlias
		{
			get
			{
				return _antiAlias;
			}
			set
			{
				if (AntiAlias != value)
				{
					_antiAlias = value;
					AdjustHeight();
				}
			}
		}

		public Color BorderColor
		{
			get
			{
				return _borderColor;
			}
			set
			{
				if (BorderColor != value)
				{
					_borderColor = value;
					Invalidate();
				}
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
		public Borders Borders
		{
			get
			{
				return _borders;
			}
		}

		[DefaultValue(1)]
		private int BorderWidth
		{
			get
			{
				return 1;
			}
		}

		public override Cursor Cursor
		{
			get
			{
				return base.Cursor;
			}
			set
			{
				base.Cursor = value;
			}
		}

		protected virtual string FullText
		{
			get
			{
				return Text;
			}
		}

		[DefaultValue(-1)]
		[Editor("System.Windows.Forms.Design.ImageIndexEditor, System.Design",
			 "System.Drawing.Design.UITypeEditor, System.Drawing")]
		[TypeConverter("System.Windows.Forms.ImageIndexConverter, System.Windows.Forms")]
		public int ImageIndex
		{
			get
			{
				return _imageIndex;
			}
			set
			{
				_imageIndex = value;
				Invalidate();
			}
		}

		public ImageList ImageList
		{
			get
			{
				return _imageList;
			}
			set
			{
				if (ImageList != null)
				{
					ImageList.Disposed -= new EventHandler(ImageListDisposed);
					ImageList.RecreateHandle -= new EventHandler(ImageListRecreateHandle);
				}
				_imageList = value;
				if (ImageList != null)
				{
					ImageList.Disposed += new EventHandler(ImageListDisposed);
					ImageList.RecreateHandle += new EventHandler(ImageListRecreateHandle);
				}
				AdjustHeight();
			}
		}

		[DefaultValue(4)]
		public int ImagePadding
		{
			get
			{
				return _imagePadding;
			}
			set
			{
				if (ImagePadding != value)
				{
					_imagePadding = value;
					AdjustHeight();
				}
			}
		}

		protected Rectangle ImageHitTestRectangle
		{
			get
			{
				if (ImageList != null && ImageIndex >= 0)
				{
					return new Rectangle(Padding.Left, Padding.Top,
						ImageList.ImageSize.Width, ImageList.ImageSize.Height);
				}
				else
					return Rectangle.Empty;
			}
		}

		protected Rectangle ImagePaddingHitTestRectangle
		{
			get
			{
				Rectangle rect = ImageHitTestRectangle;
				if (!rect.IsEmpty)
				{
					rect.X = rect.Right;
					rect.Width = ImagePadding;
				}
				return rect;
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
		public Padding Padding
		{
			get
			{
				return _padding;
			}
		}

		protected Rectangle TextRectangleOuter
		{
			get
			{
				Padding padding = AllTextPadding;
				return Rectangle.FromLTRB(padding.Left, padding.Top,
					Width - padding.Right, Height - padding.Bottom);
			}
		}

		private int TotalUnderbarHeight
		{
			get
			{
				return UnderbarHeight + 1;
			}
		}

		[DefaultValue(false)]
		public bool Underbar
		{
			get
			{
				return _underbar;
			}
			set
			{
				if (Underbar != value)
				{
					_underbar = value;
					AdjustHeight();
				}
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		[Obsolete("Use Underbar instead. Underline will be removed in a future version of SharpLayout.")]
		public bool Underline
		{
			get
			{
				return Underbar;
			}
			set
			{
				Underbar = value;
			}
		}

		private int UnderbarHeight
		{
			get
			{
				return 1;
			}
		}


		protected void AdjustHeight()
		{
			Height = GetHeight(Width);
			Invalidate();
		}

		private void BordersChanged(object sender, EventArgs e)
		{
			AdjustHeight();
		}

		private StringFormat CreateStringFormat()
		{
			StringFormat stringFormat;

			if (AntiAlias)
				stringFormat = StringFormat.GenericTypographic;
			else
				stringFormat = new StringFormat();

			stringFormat.FormatFlags &= ~StringFormatFlags.LineLimit;
			stringFormat.Trimming = StringTrimming.None;
			return stringFormat;
		}

		private int GetHeight(int width)
		{
			Padding allTextPadding = AllTextPadding;
			int textWidth = width - allTextPadding.Left - allTextPadding.Right;
			int textHeight = GetTextExtent(textWidth).Height +
				allTextPadding.Top + allTextPadding.Bottom;

			int imageHeight = 0;
			if (ImageList != null)
				imageHeight = ImageList.ImageSize.Height;
			imageHeight += AllPadding.Top + AllPadding.Bottom;

			return Math.Max(textHeight, imageHeight);
		}

		private Size GetTextExtent(int textWidth)
		{
			string flowText;

			// Make sure we have at least one character to measure
			flowText = FullText;
			if (flowText.Trim().Length == 0)
				flowText = "0";

			using (Graphics g = CreateGraphics())
			using (StringFormat stringFormat = CreateStringFormat())
			{
				SizeF extent = g.MeasureString(flowText, Font, textWidth, stringFormat);
				return new Size((int) Math.Round(extent.Width),
					(int) Math.Round(extent.Height));
			}
		}

		private void ImageListDisposed(object sender, EventArgs e)
		{
            ImageList = null;
		}

		private void ImageListRecreateHandle(object sender, EventArgs e)
		{
			AdjustHeight();
		}

		protected override void OnFontChanged(EventArgs e)
		{
			AdjustHeight();
			base.OnFontChanged(e);
		}

		protected override void OnPaint(PaintEventArgs e)
		{
			if (DesignMode)
				PaintDesignBorder(e.Graphics);
			if (Underbar)
				PaintUnderbar(e.Graphics);
			PaintBorders(e.Graphics);
			PaintImage(e.Graphics);
			PaintText(e.Graphics);
			base.OnPaint(e);
		}

		protected override void OnResize(EventArgs e)
		{
			Invalidate();
			base.OnResize(e);
		}

		protected override void OnTextChanged(EventArgs e)
		{
			AdjustHeight();
			base.OnTextChanged(e);
		}

		private void PaddingChanged(object sender, EventArgs e)
		{
			AdjustHeight();
		}

		private void PaintBorders(Graphics g)
		{
			int paintLeft = 0;
			int paintTop = 0;
			int paintRight = Width;
			int paintBottom = Height;
			using (Brush brush = new SolidBrush(BorderColor))
			{
				if (Borders.Top)
				{
					g.FillRectangle(brush, paintLeft, paintTop,
						paintRight - paintLeft, BorderWidth);
					paintTop += BorderWidth;
				}
				if (Borders.Bottom)
				{
					g.FillRectangle(brush, paintLeft, paintBottom - BorderWidth,
						paintRight - paintLeft, BorderWidth);
					paintBottom -= BorderWidth;
				}
				if (Borders.Left)
				{
					g.FillRectangle(brush, paintLeft, paintTop,
						BorderWidth, paintBottom - paintTop);
					paintLeft += BorderWidth;
				}
				if (Borders.Right)
				{
					g.FillRectangle(brush, paintRight - BorderWidth, paintTop,
						BorderWidth, paintBottom - paintTop);
				}
			}
		}

		private void PaintDesignBorder(Graphics g)
		{
			int channelValue;
			if (BackColor.GetBrightness() >= 0.5)
				channelValue = 0;
			else
				channelValue = 255;

			using (SolidBrush brush = new SolidBrush(
					   Color.FromArgb(24, channelValue, channelValue, channelValue)))
			{
				g.FillRectangle(brush, 0, 0, Width, 1);
				g.FillRectangle(brush, 0, Height - 1, Width, 1);
				g.FillRectangle(brush, 0, 1, 1, Height - 2);
				g.FillRectangle(brush, Width - 1, 1, 1, Height - 2);
			}
		}

		private void PaintImage(Graphics g)
		{
			if (ImageList == null || ImageIndex < 0)
				return;

			ImageList.Draw(g, AllPadding.Left, AllPadding.Top, ImageIndex);
		}

		private void PaintText(Graphics g)
		{
			if (AntiAlias)
				g.TextRenderingHint = TextRenderingHint.AntiAlias;
			else
				g.TextRenderingHint = TextRenderingHint.SystemDefault;

			using (StringFormat stringFormat = CreateStringFormat())
			{
				PaintTextInner(g, FullText, Font, ForeColor, TextRectangleOuter, stringFormat);
			}
		}

		protected virtual void PaintTextInner(Graphics g, string text, Font font,
			Color foreColor, Rectangle textRectangle, StringFormat stringFormat)
		{
			using (SolidBrush brush = new SolidBrush(foreColor))
			{
				g.DrawString(text, font, brush, textRectangle, stringFormat);
			}
		}

		private void PaintUnderbar(Graphics g)
		{
			int underbarBottom = Height - Padding.Bottom;
			if (Borders.Bottom)
				underbarBottom -= BorderWidth;

			Rectangle underbarRect = Rectangle.FromLTRB(Padding.Left,
				underbarBottom - UnderbarHeight,
				Width - Padding.Right, underbarBottom);
			g.FillRectangle(SystemBrushes.ControlDark, underbarRect);
		}

		protected override void SetBoundsCore(int x, int y, int width, int height, BoundsSpecified specified)
		{
			int newHeight = GetHeight(width);
			base.SetBoundsCore(x, y, width, newHeight, specified);
		}

	}
}