// Copyright  2004 by Joe White (joe@excastle.com).
// Licensed under the Artistic License. See Artistic.txt.

using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Security;
using System.Windows.Forms;

namespace SharpLayout
{
	[DefaultEvent("LinkClicked")]
	public class FlowLinkLabel : FlowLabel
	{
		private delegate void MessageHandler(Message message);

		private ContextMenu _contextMenu;
		private Color _foreColorActive = Color.Red;
		private Color _foreColorHover = Color.Red;
		private Color _foreColorLink = Color.Blue;
		private MessageHandler _handleMessage;
		private bool _isActive;
		private bool _isLeftButtonDown;
		private bool _isMouseOver;
		private bool _isRightButtonDown;
		private RectangleF[] _textHitTestRectangles;
		private string _textSuffix;

		public FlowLinkLabel()
		{
			_handleMessage = new MessageHandler(HandleMessageDefault);
			SetStyle(ControlStyles.EnableNotifyMessage, true);
		}


		public new ContextMenu ContextMenu
		{
			get
			{
				return _contextMenu;
			}
			set
			{
				_contextMenu = value;
			}
		}

		public override Cursor Cursor
		{
			get
			{
				if (!DesignMode)
				{
					FlowLabelHitTest hitTest = GetHitTest(PointToClient(Cursor.Position));
					if (hitTest != FlowLabelHitTest.None)
						return Cursors.Hand;
				}
				return base.Cursor;
			}
			set
			{
				base.Cursor = value;
			}
		}

		[DefaultValue(typeof(Color), "Red")]
		public Color ForeColorActive
		{
			get
			{
				return _foreColorActive;
			}
			set
			{
				_foreColorActive = value;
				Invalidate();
			}
		}

		[DefaultValue(typeof(Color), "Red")]
		public Color ForeColorHover
		{
			get
			{
				return _foreColorHover;
			}
			set
			{
				_foreColorHover = value;
				Invalidate();
			}
		}

		[DefaultValue(typeof(Color), "Blue")]
		public Color ForeColorLink
		{
			get
			{
				return _foreColorLink;
			}
			set
			{
				_foreColorLink = value;
				Invalidate();
			}
		}

		protected override string FullText
		{
			get
			{
				return Text + TextSuffix;
			}
		}

		private bool IsActive
		{
			get
			{
				return _isActive;
			}
			set
			{
				if (IsActive != value)
				{
					_isActive = value;
					Refresh();
				}
			}
		}

		private bool IsLeftButtonDown
		{
			get
			{
				return _isLeftButtonDown;
			}
			set
			{
				if (IsLeftButtonDown != value)
				{
					_isLeftButtonDown = value;
					Refresh();
				}
			}
		}

		private bool IsMouseOver
		{
			get
			{
				return _isMouseOver;
			}
			set
			{
				if (IsMouseOver != value)
				{
					_isMouseOver = value;
					Refresh();
				}
			}
		}

		private bool IsRightButtonDown
		{
			get
			{
				return _isRightButtonDown;
			}
			set
			{
				_isRightButtonDown = value;
			}
		}

		public string TextSuffix
		{
			get
			{
				return _textSuffix;
			}
			set
			{
				_textSuffix = value;
				_textHitTestRectangles = null;
				AdjustHeight();
			}
		}

		[Browsable(false)]
		public new event EventHandler Click
		{
			add
			{
				base.Click += value;
			}
			remove
			{
				base.Click -= value;
			}
		}

		public event EventHandler LinkClicked;

		/// <summary>
		/// Creates a new Region representing the text hit-test area.
		/// The caller is responsible for calling Dispose on this Region.
		/// </summary>
		/// <returns>The new Region.</returns>
		private Region CreateTextRegion()
		{
			TextHitTestRectanglesNeeded();

			Region region = null;
			foreach (RectangleF rect in _textHitTestRectangles)
			{
				Region newRegion = new Region(rect);
				if (region == null)
					region = newRegion;
				else
				{
					region.Union(newRegion);
					newRegion.Dispose();
				}
			}

			return region;
		}

		public FlowLabelHitTest GetHitTest(Point point)
		{
			TextHitTestRectanglesNeeded();

			foreach (RectangleF rect in _textHitTestRectangles)
			{
				if (rect.Contains(point))
					return FlowLabelHitTest.Text;
			}

			if (ImageHitTestRectangle.Contains(point))
				return FlowLabelHitTest.Image;
			else if (ImagePaddingHitTestRectangle.Contains(point))
				return FlowLabelHitTest.ImagePadding;
			else
				return FlowLabelHitTest.None;
		}

		private void HandleMessageDefault(Message message)
		{
			// This is the first time _handleMessage has been invoked.
			// See if we can call HandleMessageTrusted().
			try
			{
				// The JIT will throw an exception if we don't have enough permissions.
				HandleMessageTrusted(message);

				// Success. Short-circuit past this check (and its try..finally) in the future.
				_handleMessage = new MessageHandler(HandleMessageTrusted);
			}
			catch (SecurityException)
			{
				// We're not trusted code, so we can't access message's members.
				// Fall back to reduced functionality in the future; in particular,
				// we won't be able to detect WM_CANCELMODE.
				_handleMessage = null;
			}
		}

		private void HandleMessageTrusted(Message message)
		{
			const int WM_CANCELMODE = 0x001F;

			if (message.Msg == WM_CANCELMODE)
				WMCancelMode();
		}

		private bool IsEventInHotArea(MouseEventArgs e)
		{
			Point point = new Point(e.X, e.Y);
			FlowLabelHitTest hitTest = GetHitTest(point);
			return (hitTest != FlowLabelHitTest.None);
		}

		private void OnLinkClicked()
		{
			IsActive = true;
			try
			{
				EventHandler linkClicked = LinkClicked;
				if (linkClicked != null)
					linkClicked(this, EventArgs.Empty);
			}
			finally
			{
				IsActive = false;
			}
		}

		protected override void OnMouseDown(MouseEventArgs e)
		{
			if (IsEventInHotArea(e))
			{
				if (e.Button == MouseButtons.Left)
					IsLeftButtonDown = true;
				else if (e.Button == MouseButtons.Right)
					IsRightButtonDown = true;
			}

			base.OnMouseDown(e);
		}

		protected override void OnMouseLeave(EventArgs e)
		{
			IsMouseOver = false;
			base.OnMouseLeave(e);
		}

		protected override void OnMouseMove(MouseEventArgs e)
		{
			IsMouseOver = IsEventInHotArea(e);
			base.OnMouseMove(e);
		}

		protected override void OnMouseUp(MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				if (IsLeftButtonDown && IsEventInHotArea(e))
					OnLinkClicked();

				IsLeftButtonDown = false;
			}
			else if (e.Button == MouseButtons.Right)
			{
				if (IsRightButtonDown && IsEventInHotArea(e))
					ShowContextMenu(new Point(e.X, e.Y));

				IsRightButtonDown = false;
			}

			base.OnMouseUp(e);
		}

		protected override void OnNotifyMessage(Message m)
		{
			if (_handleMessage != null)
				_handleMessage(m);

			base.OnNotifyMessage(m);
		}

		protected override void OnResize(EventArgs e)
		{
			_textHitTestRectangles = null;
			base.OnResize(e);
		}

		protected override void OnTextChanged(EventArgs e)
		{
			_textHitTestRectangles = null;
			base.OnTextChanged(e);
		}

		protected override void PaintTextInner(Graphics g, string text, Font font,
			Color foreColor, Rectangle textRectangle, StringFormat stringFormat)
		{
			Color linkColor;
			if (IsActive)
				linkColor = ForeColorActive;
			else if (IsMouseOver)
			{
				if (IsLeftButtonDown)
					linkColor = ForeColorActive;
				else
					linkColor = ForeColorHover;
			}
			else
				linkColor = ForeColorLink;

			using (Font underlinedFont = new Font(font, font.Style | FontStyle.Underline))
			using (Region textRegion = CreateTextRegion())
			{
				GraphicsContainer container;
				
				container = g.BeginContainer();
				g.SetClip(textRegion, CombineMode.Intersect);
				base.PaintTextInner(g, text, underlinedFont, linkColor, textRectangle, stringFormat);
				g.EndContainer(container);

				container = g.BeginContainer();
				textRegion.Complement(ClientRectangle);
				g.SetClip(textRegion, CombineMode.Intersect);
				base.PaintTextInner(g, text, Font, ForeColor, textRectangle, stringFormat);
				g.EndContainer(container);
			}
		}

		private void ShowContextMenu(Point point)
		{
			IsActive = true;
			try
			{
				ContextMenu.Show(this, point);
			}
			finally
			{
				IsActive = false;
			}
		}

		private void TextHitTestRectanglesNeeded()
		{
			if (_textHitTestRectangles != null)
				return;

			using (Graphics g = this.CreateGraphics())
			using (Matrix matrix = new Matrix())
			using (StringFormat stringFormat = new StringFormat())
			{
				CharacterRange range = new CharacterRange(0, Text.Length);
				stringFormat.SetMeasurableCharacterRanges(new CharacterRange[] { range });

				Region[] regions = g.MeasureCharacterRanges(FullText, Font,
					TextRectangleOuter, stringFormat);
				_textHitTestRectangles = regions[0].GetRegionScans(matrix);
			}

			// Now adjust for the blank pixels between lines, and the unexplained
			// blank pixels at the left side.
			// Note that we can compensate for the left-side pixels only because
			// we know that the hyperlink text is always at the beginning of the
			// string and the text is always left-justified, therefore every
			// rectangle will be starting at the left margin.
			int textLeft = TextRectangleOuter.Left;
			for (int i = 0; i < _textHitTestRectangles.Length; ++i)
			{
				RectangleF rect = _textHitTestRectangles[i];

				float newLeft = rect.Left;
				if (newLeft > textLeft)
					newLeft = textLeft;

				float newTop = rect.Top;
				if (i > 0)
				{
					RectangleF prevRect = _textHitTestRectangles[i - 1];
					if (newTop > prevRect.Bottom)
						newTop = prevRect.Bottom;
					newTop = prevRect.Bottom;
				}

				_textHitTestRectangles[i] = RectangleF.FromLTRB(newLeft, newTop,
					rect.Right, rect.Bottom);
			}
		}

		private void WMCancelMode()
		{
			IsLeftButtonDown = false;
			IsRightButtonDown = false;
		}

	}
}